@@ -41,12 +41,18 @@ AMQTable_SetStringValue(amqp_table_t*, amqp_bytes_t, amqp_bytes_t);
 _PYRMQ_INLINE void
 AMQTable_SetIntValue(amqp_table_t *, amqp_bytes_t, int);
 
+_PYRMQ_INLINE void
+AMQTable_SetNilValue(amqp_table_t *, amqp_bytes_t);
+
 _PYRMQ_INLINE void
 AMQTable_SetDoubleValue(amqp_table_t *, amqp_bytes_t, double);
 
 _PYRMQ_INLINE amqp_field_value_t*
 AMQArray_AddEntry(amqp_array_t*);
 
+_PYRMQ_INLINE void
+AMQArray_SetNilValue(amqp_array_t*);
+
 _PYRMQ_INLINE void
 AMQArray_SetTableValue(amqp_array_t*, amqp_table_t);
 
@@ -99,7 +105,7 @@ void PyRabbitMQ_SetErr_UnexpectedHeader(amqp_frame_t*);
 int PyRabbitMQ_Not_Connected(PyRabbitMQ_Connection *);
 
 static amqp_table_t PyDict_ToAMQTable(amqp_connection_state_t, PyObject *, amqp_pool_t *);
-static amqp_array_t PyList_ToAMQArray(amqp_connection_state_t, PyObject *, amqp_pool_t *);
+static amqp_array_t PyIter_ToAMQArray(amqp_connection_state_t, PyObject *, amqp_pool_t *);
 
 static PyObject* AMQTable_toPyDict(amqp_table_t *table);
 static PyObject* AMQArray_toPyList(amqp_array_t *array);
@@ -163,6 +169,12 @@ AMQTable_SetIntValue(amqp_table_t *table,
     entry->value.value.i32 = value;
 }
 
+_PYRMQ_INLINE void
+AMQTable_SetNilValue(amqp_table_t *table, amqp_bytes_t key) {
+    amqp_table_entry_t *entry = AMQTable_AddEntry(table, key);
+    entry->value.kind = AMQP_FIELD_KIND_VOID;
+}
+
 _PYRMQ_INLINE void
 AMQTable_SetDoubleValue(amqp_table_t *table,
                         amqp_bytes_t key, double value)
@@ -180,6 +192,13 @@ AMQArray_AddEntry(amqp_array_t *array)
     return entry;
 }
 
+_PYRMQ_INLINE void
+AMQArray_SetNilValue(amqp_array_t *array)
+{
+    amqp_field_value_t *entry = AMQArray_AddEntry(array);
+    entry->kind = AMQP_FIELD_KIND_VOID;
+}
+
 _PYRMQ_INLINE void
 AMQArray_SetTableValue(amqp_array_t *array, amqp_table_t value)
 {
@@ -231,17 +250,21 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool
     dst.entries = amqp_pool_alloc(pool, size * sizeof(amqp_table_entry_t));
     while (PyDict_Next(src, &pos, &dkey, &dvalue)) {
 
-        if (PyDict_Check(dvalue)) {
+        if (dvalue == Py_None) {
+            /* None */
+            AMQTable_SetNilValue(&dst, PyString_AS_AMQBYTES(dkey));
+        }
+        else if (PyDict_Check(dvalue)) {
             /* Dict */
             AMQTable_SetTableValue(&dst,
                     PyString_AS_AMQBYTES(dkey),
                     PyDict_ToAMQTable(conn, dvalue, pool));
         }
-        else if (PyList_Check(dvalue)) {
+        else if (PyList_Check(dvalue) || PyTuple_Check(dvalue)) {
             /* List */
             AMQTable_SetArrayValue(&dst,
                     PyString_AS_AMQBYTES(dkey),
-                    PyList_ToAMQArray(conn, dvalue, pool));
+                    PyIter_ToAMQArray(conn, dvalue, pool));
         }
         else if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
             /* Int | Long */
@@ -253,6 +276,15 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool
                     clong_value
             );
         }
+        else if (PyFloat_Check(dvalue)) {
+            cdouble_value = PyFloat_AsDouble(dvalue);
+            if (PyErr_Occurred())
+                goto error;
+            AMQTable_SetDoubleValue(&dst,
+                PyString_AS_AMQBYTES(dkey),
+                    cdouble_value
+            );
+        }
         else {
             /* String | Unicode */
             is_unicode = PyUnicode_Check(dvalue);
@@ -267,23 +299,11 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool
                 );
             }
             else {
-                cdouble_value = PyFloat_AsDouble(dvalue);
-                if (PyErr_Occurred())
-                    goto error;
-                if (PyFloat_Check(dvalue)) {
-
-                    AMQTable_SetDoubleValue(&dst,
-                        PyString_AS_AMQBYTES(dkey),
-                        cdouble_value
-                    );
-                }
-                else {
-                    /* unsupported type */
-                    PyErr_Format(PyExc_ValueError,
-                        "Table member %s is of an unsupported type",
-                        PyString_AS_STRING(dkey));
-                    goto error;
-            }
+                /* unsupported type */
+                PyErr_Format(PyExc_ValueError,
+                    "Table member %s is of an unsupported type",
+                    PyString_AS_STRING(dkey));
+                goto error;
             }
         }
     }
@@ -294,63 +314,73 @@ error:
 }
 
 static amqp_array_t
-PyList_ToAMQArray(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool)
+PyIter_ToAMQArray(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool)
 {
-    PyObject *dvalue = NULL;
-    PY_SIZE_TYPE size = 0;
     PY_SIZE_TYPE pos = 0;
     uint64_t clong_value = 0;
     int is_unicode = 0;
     amqp_array_t dst = AMQP_EMPTY_ARRAY;
 
-    size = PyList_Size(src);
+    PY_SIZE_TYPE size = PySequence_Size(src);
+    if (size == -1) return dst;
 
-    /* allocate new array */
+    PyObject *iterator = PyObject_GetIter(src);
+    if (iterator == NULL) return dst;
+    PyObject *item = NULL;
+
+    /* allocate new amqp array */
     dst.num_entries = 0;
     dst.entries = amqp_pool_alloc(pool, size * sizeof(amqp_field_value_t));
-    for (pos = 0; pos < size; ++pos) {
 
-        dvalue = PyList_GetItem(src, pos);
-
-        if (PyDict_Check(dvalue)) {
+    while (item = PyIter_Next(iterator)) {
+        if (item == Py_None) {
+            /* None */
+            AMQArray_SetNilValue(&dst);
+        }
+        else if (PyDict_Check(item)) {
             /* Dict */
-            AMQArray_SetTableValue(&dst,
-                    PyDict_ToAMQTable(conn, dvalue, pool));
+            AMQArray_SetTableValue(
+                &dst, PyDict_ToAMQTable(conn, item, pool));
         }
-        else if (PyList_Check(dvalue)) {
+        else if (PyList_Check(item) || PyTuple_Check(item)) {
             /* List */
-            AMQArray_SetArrayValue(&dst,
-                    PyList_ToAMQArray(conn, dvalue, pool));
+            AMQArray_SetArrayValue(
+                &dst, PyIter_ToAMQArray(conn, item, pool));
         }
-        else if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
+        else if (PyLong_Check(item) || PyInt_Check(item)) {
             /* Int | Long */
-            clong_value = (int64_t)PyLong_AsLong(dvalue);
+            clong_value = (int64_t)PyLong_AsLong(item);
             AMQArray_SetIntValue(&dst, clong_value);
         }
         else {
             /* String | Unicode */
-            is_unicode = PyUnicode_Check(dvalue);
-            if (is_unicode || PyString_Check(dvalue)) {
+            is_unicode = PyUnicode_Check(item);
+            if (is_unicode || PyString_Check(item)) {
                 if (is_unicode) {
-                    if ((dvalue = PyUnicode_AsASCIIString(dvalue)) == NULL)
-                        goto error;
+                    if ((item = PyUnicode_AsASCIIString(item)) == NULL)
+                        goto item_error;
                 }
-                AMQArray_SetStringValue(&dst,
-                        PyString_AS_AMQBYTES(dvalue));
+                AMQArray_SetStringValue(
+                    &dst, PyString_AS_AMQBYTES(item));
             }
             else {
                 /* unsupported type */
                 PyErr_Format(PyExc_ValueError,
                     "Array member at index %lu, %s, is of an unsupported type",
-                    pos, PyObject_REPR(dvalue));
-                goto error;
+                    pos, PyObject_REPR(item));
+                goto item_error;
             }
         }
+        Py_XDECREF(item);
     }
+
     return dst;
+item_error:
+    Py_XDECREF(item);
 error:
+    Py_XDECREF(iterator);
     assert(PyErr_Occurred());
-    return AMQP_EMPTY_ARRAY;
+    return dst;
 }
 
 
@@ -481,6 +511,9 @@ AMQTable_toPyDict(amqp_table_t *table)
         int i;
         for (i = 0; i < table->num_entries; ++i, key=value=NULL) {
             switch (table->entries[i].value.kind) {
+                case AMQP_FIELD_KIND_VOID:
+                    value = Py_None;
+                    break;
                 case AMQP_FIELD_KIND_BOOLEAN:
                     value = PyBool_FromLong(AMQTable_VAL(table, i, boolean));
                     break;
@@ -531,11 +564,11 @@ AMQTable_toPyDict(amqp_table_t *table)
             }
 
             key = AMQTable_TO_PYKEY(table, i);
-            if (value)
-                PyDICT_SETKV_DECREF(dict, key, value);
-            else
+            if (value == Py_None || value == NULL)
                 /* unsupported type */
                 PyDICT_SETNONE_DECREF(dict, key);
+            else
+                PyDICT_SETKV_DECREF(dict, key, value);
 
         }
     }
@@ -614,7 +647,6 @@ AMQArray_toPyList(amqp_array_t *array)
     return list;
 }
 
-
 _PYRMQ_INLINE int
 PyDict_to_basic_properties(PyObject *p,
                            amqp_basic_properties_t *props,
@@ -685,8 +717,7 @@ PyDict_to_basic_properties(PyObject *p,
 
     if ((value = PyDict_GetItemString(p, "headers")) != NULL) {
         props->headers = PyDict_ToAMQTable(conn, value, pool);
-        if (PyErr_Occurred())
-            return -1;
+        if (PyErr_Occurred()) return -1;
     }
     return 1;
 }
@@ -1072,7 +1103,7 @@ PyRabbitMQ_Connection_destroy_channel(PyRabbitMQ_Connection *self,
     amqp_rpc_reply_t reply;
     Py_BEGIN_ALLOW_THREADS;
     reply = amqp_channel_close(self->conn, channel, AMQP_REPLY_SUCCESS);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     return PyRabbitMQ_HandleAMQError(self, channel, reply, "Couldn't close channel");
@@ -1160,6 +1191,40 @@ finally:
     return retval;
 }
 
+void
+PyRabbitMQ_SetErr_ReceivedFrame(PyRabbitMQ_Connection *self, amqp_frame_t* frame)
+{
+    static char errstr[512];
+
+    switch(frame->payload.method.id) {
+        case AMQP_CHANNEL_CLOSE_METHOD: {
+            amqp_channel_close_t *chanm = (amqp_channel_close_t *)frame->payload.method.decoded;
+            snprintf(errstr, sizeof(errstr),
+                     "channel error %d, message: %.*s",
+                     chanm->reply_code,
+                     (int) chanm->reply_text.len,
+                     (char *) chanm->reply_text.bytes);
+            PyErr_SetString(PyRabbitMQExc_ChannelError, errstr);
+            PyRabbitMQ_revive_channel(self, frame->channel);
+            break;
+        }
+        case AMQP_CONNECTION_CLOSE_METHOD: {
+            amqp_connection_close_t *connm = (amqp_connection_close_t *)frame->payload.method.decoded;
+            snprintf(errstr, sizeof(errstr),
+                "server connection error %d message: %.*s",
+                connm->reply_code,
+                (int) connm->reply_text.len,
+                (char *) connm->reply_text.bytes);
+            PyErr_SetString(PyRabbitMQExc_ConnectionError, errstr);
+            PyRabbitMQ_Connection_close(self);
+            break;
+        }
+        default: {
+            PyErr_SetString(PyRabbitMQExc_ConnectionError, "Bad frame read");
+            break;
+        }
+    }
+}
 
 int
 PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
@@ -1191,9 +1256,8 @@ PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
             retval = amqp_simple_wait_frame(conn, &frame);
             Py_END_ALLOW_THREADS;
             if (retval < 0) break;
-            if (frame.frame_type != AMQP_FRAME_METHOD
-                    || frame.payload.method.id != AMQP_BASIC_DELIVER_METHOD)
-                continue;
+            if (frame.frame_type != AMQP_FRAME_METHOD) continue;
+            if (frame.payload.method.id != AMQP_BASIC_DELIVER_METHOD) goto altframe;
 
             delivery_info = PyDict_New();
             deliver = (amqp_basic_deliver_t *)frame.payload.method.decoded;
@@ -1217,6 +1281,7 @@ PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
         Py_END_ALLOW_THREADS;
         if (retval < 0) break;
 
+        if (frame.frame_type == AMQP_FRAME_METHOD) goto altframe;
         if (frame.frame_type != AMQP_FRAME_HEADER) {
             PyRabbitMQ_SetErr_UnexpectedHeader(&frame);
             goto finally;
@@ -1239,6 +1304,7 @@ PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
             Py_END_ALLOW_THREADS;
             if (retval < 0) break;
 
+            if (frame.frame_type == AMQP_FRAME_METHOD) goto altframe;
             if (frame.frame_type != AMQP_FRAME_BODY) {
                 PyErr_SetString(PyRabbitMQExc_ChannelError,
                     "Expected body, got unexpected frame");
@@ -1288,6 +1354,8 @@ PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
     }
     goto finally;
 
+altframe:
+    PyRabbitMQ_SetErr_ReceivedFrame(self, &frame);
 error:
     retval = -1;
 
@@ -1344,7 +1412,7 @@ PyRabbitMQ_Connection_queue_bind(PyRabbitMQ_Connection *self,
                         PyString_AS_AMQBYTES(routing_key),
                         bargs);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.bind"))
@@ -1399,7 +1467,7 @@ PyRabbitMQ_Connection_queue_unbind(PyRabbitMQ_Connection *self,
                       PyString_AS_AMQBYTES(routing_key),
                       uargs);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.unbind"))
@@ -1440,7 +1508,7 @@ PyRabbitMQ_Connection_queue_delete(PyRabbitMQ_Connection *self,
             (amqp_boolean_t)if_empty);
     if (ok == NULL)
         reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (ok == NULL && PyRabbitMQ_HandleAMQError(self, channel,
@@ -1542,7 +1610,7 @@ PyRabbitMQ_Connection_queue_purge(PyRabbitMQ_Connection *self,
     ok = amqp_queue_purge(self->conn, channel,
                           PyString_AS_AMQBYTES(queue));
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.purge"))
@@ -1599,7 +1667,7 @@ PyRabbitMQ_Connection_exchange_declare(PyRabbitMQ_Connection *self,
                           PyString_AS_AMQBYTES(type),
                           (amqp_boolean_t)passive,
                           (amqp_boolean_t)durable,
-                          eargs
+                          0, 0, eargs
     );
     reply = amqp_get_rpc_reply(self->conn);
     Py_END_ALLOW_THREADS;
@@ -1636,7 +1704,7 @@ PyRabbitMQ_Connection_exchange_delete(PyRabbitMQ_Connection *self,
                          PyString_AS_AMQBYTES(exchange),
                          (amqp_boolean_t)if_unused);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "exchange.delete"))
@@ -1682,8 +1750,9 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
 
     Py_INCREF(propdict);
     channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
-    if (!PyDict_to_basic_properties(propdict, &props, self->conn, channel_pool))
+    if (PyDict_to_basic_properties(propdict, &props, self->conn, channel_pool) < 1) {
         goto bail;
+    }
     Py_DECREF(propdict);
 
     bytes.len = (size_t)body_size;
@@ -1697,11 +1766,12 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
                              (amqp_boolean_t)immediate,
                              &props,
                              bytes);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleError(ret, "basic.publish"))
+    if (!PyRabbitMQ_HandleError(ret, "basic.publish")) {
         goto error;
+    }
     Py_RETURN_NONE;
 
 error:
@@ -1804,7 +1874,7 @@ PyRabbitMQ_Connection_basic_cancel(PyRabbitMQ_Connection *self,
     ok = amqp_basic_cancel(self->conn, channel,
                            PyString_AS_AMQBYTES(consumer_tag));
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "basic.cancel"))
@@ -1927,7 +1997,7 @@ PyRabbitMQ_Connection_flow(PyRabbitMQ_Connection *self,
     Py_BEGIN_ALLOW_THREADS;
     ok = amqp_channel_flow(self->conn, channel, (amqp_boolean_t)active);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "channel.flow"))
@@ -1960,7 +2030,7 @@ PyRabbitMQ_Connection_basic_recover(PyRabbitMQ_Connection *self,
     Py_BEGIN_ALLOW_THREADS;
     ok = amqp_basic_recover(self->conn, channel, (amqp_boolean_t)requeue);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "basic.recover"))
